/**
 * @mainpage VFIS Integral Equation Solver (Single Version).
 *
 * VFIS 
 *
 * Verified Fredholm Integral Equation Solver
 * for linear Fredholm Integral Equations of the second kind
 * (Single Version)
 *
 * Verifizierter Integralgleichungsloeser 
 * fuer lineare Fredholmsche Integralgleichungen zweiter Art
 * (Einzelversion)
 *
 * Copyright (c) 2004-2007 Markus Grimmer
 * 
 * markus.grimmer@math.uni-wuppertal.de
 *
 * VFIS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * VFIS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with VFIS; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


/**
 * @file Einzel-FIGL-Loeser.cpp
 * VFIS Integral Equation Solver central module with all central methods/functions.
 * Class definitions see header file.
 */

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

#include "dim2taylor.hpp"  // 1dim.
#include "itaylor.hpp"     // 2dim.
#include "ilss.hpp"        // LSS (Holbig)

#include "interval.hpp"
#include "ivector.hpp"
#include "intvector.hpp"
#include "imatrix.hpp"

#include "dot.hpp"
#include "idot.hpp"

#include "Einzel-FIGL-Loeser.hpp"

using namespace std;
using namespace cxsc;
using namespace taylor;

// *****************************************************************************
// *****************************************************************************
// *****************************************************************************

// -----------------------------------------------------------------------------
// Auxiliary functions
// -----------------------------------------------------------------------------

/**
 * Interval Midpoint Computation with interval result.
 * @param[in] I The interval
 * @return The midpoint
 */
interval Imid(interval& I)
{
  return interval(Inf(I))+interval(Sup(I))/real(2.0);                        
}

// -----------------------------------------------------------------------------

/**
 * Epsilon inflation for itaylor arguments (elementwise).
 * @param[in]  it itaylor object
 * @param[in]  eps Epsilon
 * @return The inflated itaylor object
 */
itaylor blow(itaylor& it, real eps)
{
  return (real(1.0)+eps)*it-eps*it;
}

// -----------------------------------------------------------------------------

/**
 * Computes powers I^j, j=1..ord, for a given interval.
 * @param[in]  baseInt interval
 * @param[in]  ord     maximum exponent
 * @return ivector containing I^j, j=1..ord
 */

ivector Powers (interval baseInt, int ord)
{
  ivector resV(0,ord);
  resV[0]=interval(1.0);
  
  for (int i=1; i<=ord; i++)
  {
    if (i%2==1) 
      resV[i]=baseInt*resV[i-1];
    else
      resV[i]=sqr(resV[i/2]);
  }
  
  return resV;
}

// -----------------------------------------------------------------------------

/**
 * Computes integrals of basic monomials.
 * @param[out]  Int_a Given a domain T:=[a,b] with midpoint t0, the integrals 
                of (a-t0)^k, k=0..intord, are computed and stored in Int_a 
 * @param[out]  Int_b Given a domain T:=[a,b] with midpoint t0, the integrals 
                of (b-t0)^k, k=0..intord, are computed and stored in Int_b 
 * @param[in]   intord Maximum order of integral computation
 * @param[in]   T domain
 */
void Monom_Integration( ivector& Int_a, ivector& Int_b, int intord, interval T )
{
  interval Tmid;
  ivector a_Pot(0,intord+1), b_Pot(0,intord+1); 
  
  Resize(Int_a,0,intord); 
  Resize(Int_b,0,intord);

  Tmid=Imid(T);
  
  a_Pot=Powers(Inf(T)-Tmid,intord+1); 
  b_Pot=Powers(Sup(T)-Tmid,intord+1); 

  for (int i=0; i<=intord; i++)
  {
    Int_a[i]=a_Pot[i+1]/real(i+1);  
    Int_b[i]=b_Pot[i+1]/real(i+1);  
  }
}

// *****************************************************************************

// -----------------------------------------------------------------------------
// class IntegralOperatorRG:
//
// see Header
// -----------------------------------------------------------------------------

/**
 * Function call operator: Applies integral operator to the function fkt
 * @param[in]  fkt The function the integral operator is applied to
 * @param[in]  ord Taylor order in use
 * @param[in]  S   Domain of the function
 * @param[in]  Int_a Integrals of powers of basic monomials
 * @param[in]  Int_b Integrals of powers of basic monomials
 * @return     itaylor representation of the result
 * @see        Monom_Integration()
 */
itaylor IntegralOperatorRG::operator() (itaylor& fkt, int ord, interval S, 
                                        ivector& Int_a, ivector& Int_b)
{
  itaylor erg;
  dotprecision accu;
  accu=real(0.0);
  idotprecision iaccu(accu,accu);
  ivector H1(0,ord), H2(0,ord), resv(0,ord);
  
  interval SmS0; 
  SmS0 = S-Imid(S);

  interval Itmp1, Itmp2;
  Itmp1=K[ord];
  Itmp2=K[ord+1];

  // Result at index [ord]:
  
  for (int i=0; i<=ord; i++)
    {
      H1[i]=Itmp1*fkt[i];
      H2[i]=Itmp2*fkt[i]*SmS0;

      accumulate(iaccu,H1[i],Int_b[i+1]);   // Genaue Variante der untenstehenden
      accumulate(iaccu,-H1[i],Int_a[i+1]);  // Berechnung (**a)
      accumulate(iaccu,H2[i],Int_b[i]);
      accumulate(iaccu,-H2[i],Int_a[i]);
    }

  rnd(iaccu, resv[ord]);
  
  // Results for other indices:

  for (int i=0; i<ord; i++)
  {
    Itmp1=K[i];
    iaccu=accu;
    //Sum=interval(0.0,0.0); // For non-accu computation
    for (int j=0; j<=ord; j++)
    {
      H1[j]=Itmp1*fkt[j];
      accumulate(iaccu,H1[j],Int_b[ord+1-i+j]);  //
      accumulate(iaccu,-H1[j],Int_a[ord+1-i+j]); // Accu version of (**b)
      //Sum+=H1[j]*Int_b[ord+1-i+j]-H1[j]*Int_a[ord+1-i+j]; (**b)
    }

    rnd(iaccu,resv[i]);
    //resv[i]=Sum;
  }

  erg=resv;

  return erg;
}

// *****************************************************************************

/**
 * Part I: Iteration function: Carries out the fixed point iteration
 * 
 * y_0=g; y_{i+1}=IntOp(y_i)
 * 
 * for an integral operator IntOp
 *
 * @param[in]  IntOp   The integral operator of the iteration
 * @param[in]  fkt_g   g as in formula above
 * @param[in]  ord     Taylor order in use
 * @param[in]  S       Domain of the function
 * @param[in]  Int_a   Integrals of powers of basic monomials
 * @param[in]  Int_b   Integrals of powers of basic monomials
 * @param[out] err     Error code
 * @param[in,out] outf Output file
 * @return     itaylor representation of the result
 */
itaylor Iteration (IntegralOperatorRG& IntOp, itaylor& fkt_g, int ord, interval S, 
                   ivector& Int_a, ivector& Int_b, int& err, ostream& outf)

{
  const real eps=1E-10;
  int itercount=0;
 
  itaylor falt, fneu;

  fneu=fkt_g;
  do
  {
    itercount++;
    falt=blow(fneu,eps);
    fneu=fkt_g+IntOp(falt, ord, S, Int_a, Int_b);
  } while ((!(fneu<=falt))&&(itercount<=10));
  
  if(itercount>10)
  {
    outf << "Iteration: Keine Einschliessung!" << endl;
    err=1;
  }
  else
  {
    outf << "Iteration: Einschliessung nach " << itercount << " Schritten." << endl;
    err=0;    
  }
  
  return fneu;
}

// *****************************************************************************
// *****************************************************************************
// *****************************************************************************


// -----------------------------------------------------------------------------
// class IntegralGleichung
//
// see Header
// -----------------------------------------------------------------------------

/**
 * Constructor for class IntegralGleichung
 * @param[in]  ordng Taylor order of functions
 */
IntegralGleichung::IntegralGleichung(int ordng)
{
  ord=ordng;
  KernFkt=0;
  RechtsFkt=0;
  Kpoint=dim2taylor(ord+1);
  Kint=dim2taylor(ord+1);
  Kern=dim2taylor(ord+1);
  RSpoint=itaylor(ord);
  RSint=itaylor(ord);
  RS=itaylor(ord);
  S=interval(0.0);
  T=interval(0.0);
  RS_S=interval(0.0);
  Smid=interval(0.0);
  Tmid=interval(0.0);
  RS_Smid=interval(0.0);
  MonomInt_a=ivector(ord);
  MonomInt_b=ivector(ord);
  Restkern=ivector(0,ord+1);
  Lsg=itaylor(ord);
  ILGS_Lsg=itaylor(ord);
}

// -----------------------------------------------------------------------------

/**
 * Constructor for class IntegralGleichung
 * @param[in]  K     Kernel function (function pointer)
 * @param[in]  KS    Domain of the first variable of K 
 * @param[in]  KT    Domain of the second variable of K 
 * @param[in]  ReS   Right hand side function (function pointer)
 * @param[in]  GS    Domain of right hand side (expected to be =KS)
 * @param[in]  ordng Taylor order of functions
 */
IntegralGleichung::IntegralGleichung(KFkt K,interval KS, interval KT,
                                     ReFkt ReS, interval GS, int ordng)
{
  KernFkt=K;
  RechtsFkt=ReS;
  S=KS;
  T=KT;
  RS_S=GS;
  Smid=Imid(S);
  Tmid=Imid(T);
  RS_Smid=Imid(RS_S);
  ord=ordng;
  Lsg=itaylor(ord);
  ILGS_Lsg=itaylor(ord);

  const int guard=3;
  int intord=2*ord+guard;

  Monom_Integration(MonomInt_a, MonomInt_b, intord, T); 
  
  // Taylor expansion kernel:

  // (K.1) "Point" Expansion for interval midpoint
  
  ivector KParam(2);
  KParam[1]=Smid;
  KParam[2]=Tmid;
  dim2taylor_vector tv;
  tv=init_var(ord+1,KParam);
  Kpoint=KernFkt(tv);

  // (K.2) Interval expansion for domain interval (for remainder)

  KParam[1]=S;
  KParam[2]=T;
  tv=init_var(ord+1,KParam);
  Kint=KernFkt(tv);
  
  // (K.3) Put them together
  
  Restkern=ivector(0,ord+1);
  
  Kern=Kpoint;
  for (int i=0; i<=ord+1; i++)
  {
    Kern[i][ord+1-i]=Kint[i][ord+1-i];
    Restkern[i]=Kint[i][ord+1-i];
  }

  // Taylor expansion right hand side
  
  // (R.1/2) "point" and interval expansion
  
  itaylor x(ord,RS_S);  
  itaylor xmid(ord,RS_Smid);
  RSint=RechtsFkt(x);   
  RSpoint=RechtsFkt(xmid);
  
  // (R.3) Put them together
  
  RS=RSpoint;
  RS[ord]=RSint[ord];
}                                     

// -----------------------------------------------------------------------------

/**
 * Part II/III of the solution algorithm: Compute the elements of the linear system
 * and solve the system (the latter with the appropriate external function).
 * 
 * @param[in]  A       Vector of iteration results
 * @param[in]  F       Iteration result of single iteration
 * @param[out] errc    Error code
 * @param[in,out] outf Output file
 */
void IntegralGleichung::Loesungsansatz(vector<itaylor>& A, itaylor& F, int& errc, ostream& outf)
{
  imatrix ILGS_Matrix(0,ord,0,ord);
  ivector ILGS_RS(0,ord);
  ivector ILGS_LsgV(0,ord);
  interval Summe, B_mmu, INull;
  ivector H1(0,ord), H2(0,ord);
  dim2taylor& B=Kern; 
  itaylor& G=F;       
  int err=0;
  
  INull=interval(0.0,0.0);
  
  // -------------------------------
  // Compute linear system entries
  // -------------------------------
  
  outf << "Berechne Eintraege des linearen Gleichungssystems..." << endl;
                  
  for (int m=0; m<=ord; m++)
  {
    for (int n=0; n<=ord; n++)
    {
      // Entry (m,n):
      
      Summe=INull;
      for (int mu=0; mu<=ord-m; mu++)  
      {
        B_mmu=B[m][mu];
        for (int nu=0; nu<=ord; nu++)
        {
          H1[nu]=B_mmu*A[n][nu];
          Summe+=H1[nu]*MonomInt_b[mu+nu]-H1[nu]*MonomInt_a[mu+nu];
        }//nu
      }//mu
      
      ILGS_Matrix[m][n]=Summe;
      
    }//n
    
    // Right hand side (Row m):
    
    Summe=INull;
    for (int mu=0; mu<=ord-m; mu++)
    {
      B_mmu=B[m][mu];
      for (int nu=0; nu<=ord; nu++)
      {
        H1[nu]=B_mmu*G[nu];
        Summe+=H1[nu]*MonomInt_b[mu+nu]-H1[nu]*MonomInt_a[mu+nu];
      }
      ILGS_RS[m]=Summe;
    }
  }//m

  // -------------------------------
  // Solve linear system
  // -------------------------------
  
  outf << "Loese lineares Gleichungssystem..." << endl;
                  
  ILGS_Matrix=ILGS_Matrix*real(-1.0);
  for (int i=0; i<=ord; i++)
    ILGS_Matrix[i][i] += real(1.0);

  LSS(ILGS_Matrix,ILGS_RS,ILGS_LsgV,err); // Solver by Holbig/Kraemer
                                          
  ILGS_Lsg=ILGS_LsgV;
  
}

// -----------------------------------------------------------------------------

/**
 * Part IV of the solution algorithm: Compute the solution from the linear
 * system solution.
 * 
 * @param[in]  A    Vector of iteration results
 * @param[in]  G    Iteration result of single iteration
 * @param[out] errc Error code
 */

void IntegralGleichung::Loesung_bilden(vector<itaylor>& A, itaylor& G, int& errc)
{
  interval INull, Summe;
  ivector H1(0,ord);
  
  INull=interval(0.0,0.0);

  if (errc==0)
  {
    for (int m=0; m<=ord; m++)
    {
      Summe=G[m];
      for (int n=0; n<=ord; n++)
        Summe+=A[n][m]*ILGS_Lsg[n];
      H1[m]=Summe;
    }
  }
  else
  {
    for (int m=0; m<=ord; m++)
      H1[m]=INull;
  }
  Lsg=H1;

}

// -----------------------------------------------------------------------------

/**
 * Main solution method containing method/subroutine calls for all parts of the 
 * integral equation solution method.
 * 
 * @param[in,out] outf Output file
 */
void IntegralGleichung::Einschliessung_berechnen(ostream& outf)
{
  itaylor F;
  int err=0,errc=0;
  vector<itaylor> Alpha(ord+1,itaylor(ord,real(0.0))); //0..ord
  // Remark: Attention with taylor constructors: They don't 
  // construct a zero vector, but (0,1,0,...)
  
  IntegralOperatorRG IRK(Restkern);

  outf << "Starte Einschliessungsberechnung..." << endl;
  
  outf << "Starte Iterationsphase..." << endl;
                  
  // -------------------------------
  // Iterations...
  // -------------------------------

  outf << "Starte Iteration F..." << endl;
  
  F=Iteration(IRK,RS,ord,RS_S,MonomInt_a,MonomInt_b,err,outf);
  
  ivector Nullv(0,ord);
  Nullv=interval(0.0,0.0);
  itaylor ITEins; 
  ITEins=Nullv;

  int m=0;
  while((m<=ord)&&(err==0))
  {
    ITEins[m]=interval(1.0,1.0);
    if (m>0)
      ITEins[m-1]=interval(0.0,0.0);
    
    outf << "Starte Iteration Alpha[" << m << "]..." << endl;
    
    Alpha[m]=Iteration(IRK,ITEins,ord,interval(0.0,0.0),MonomInt_a,MonomInt_b,errc,outf);
    m++;
    err+=errc;
  }
  
  // -------------------------------
  // Compute linear system entries
  // and solve system
  // -------------------------------
  
  if (err==0)
  {
    outf << "Berechne Loesungsansatz..."<< endl;
                    
    Loesungsansatz(Alpha,F,errc,outf);
    err+=errc;
  }
  else
  {
    outf << "Iteratives Verfahren abgebrochen bei " << m-1 << "." << endl;
  }
  
  // -------------------------------
  // Compute integral equation 
  // solution
  // -------------------------------

  outf << "Berechne Loesung der Integralgleichung... " << endl;
  
  Loesung_bilden(Alpha,F,err); 
  
  outf << "Endergebnis: " << endl;
  outf << SetPrecision(25,15) << Lsg; 

}

// -----------------------------------------------------------------------------

/**
 * Export solution function in destination system's source code.
 * 
 * @param[in]  dest     Destination system (to date, only the string "Maple"
                        is supported)
 * @param[in]  fname    File name for code output 
 * @param[in]  varname  Variable prefix for function code
 */
void IntegralGleichung::Export(string dest, string fname, string varname)
{
  ofstream expfile(fname.c_str());
  
  if (dest=="Maple")
  {
     expfile << varname << " := x -> " << Lsg[0];
     for (int j=1; j<=Lsg.order(); j++)
     {
       expfile << " &+ ( " << Lsg[j] 
               << " &* ( ( x &- " << Smid << " ) &intpower " << j << " ) ) ";
     }
     expfile << ";" << endl;
  }
  else
  {
    cerr << "Export fehlgeschlagen: Zielsystem unbekannt."  << endl;
  }
}

// -----------------------------------------------------------------------------

/**
 * Export solution function and interval plot in destination system's 
 * source code.
 * 
 * @param[in]  dest      Destination system (to date, only the string "Maple"
                         is supported)
 * @param[in]  fname     File name for code output 
 * @param[in]  steps     Number of steps in function evaluation for plot
 * @param[in]  varname   Variable prefix for function code
 * @param[in]  plottitle Title for the plot
 */
void IntegralGleichung::Export_Plot(string dest, string fname, 
                                           int steps, string varname,  
                                           string plottitle)
{
  ofstream expfile(fname.c_str());

  interval x, val;
  real xstep;    
  
  if (dest=="Maple")
  {
    expfile << varname << "_Args := []:" << endl;
    expfile << varname << "_Vals := []:" << endl;
 
    xstep=(Sup(S)-Inf(S))/real(steps);

    expfile << varname << " := x -> " << Lsg[0];
    for (int j=1; j<=Lsg.order(); j++)
    {
      expfile << " &+ ( " << Lsg[j] 
              << " &* ( ( x &- " << Smid
              << " ) &intpower " << j << " ) ) ";
    }
    expfile << ";" << endl;

    for (int j=0; j<steps; j++)
    {
      if (j<(steps-1))
        x=interval(Inf(S)+real(j)*xstep,Inf(S)+real(j+1)*xstep);
      else
        x=interval(Inf(S)+real(j)*xstep,Sup(S));

      val=Lsg(x-interval(Smid));

      expfile << varname << "_Args := [ op( " << varname << "_Args), "
              << x << " ]:" << endl;
      expfile << varname << "_Vals := [ op( " << varname << "_Vals), "
              << val << " ]:" << endl;
    }//j
    
    expfile << varname << "_Plot := interval_list_plot( "
            << varname << "_Args, " << varname << "_Vals,"
            << " color=yellow, style=patch):" << endl;
    expfile << "display(" << varname 
             << "_Plot, axes=framed, title=\"" << plottitle << "\");" << endl;
  }
  else
  {
    cerr << "Export fehlgeschlagen: Zielsystem unbekannt."  << endl;
  }
}

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

/**
 * Output operator for class IntegralGleichung objects
 * 
 * @param[in,out]  os   Output stream
 * @param[in]      IGL  IntegralGleichung object reference
 * @return         Output stream
 */
ostream& operator<< (ostream& os, IntegralGleichung& IGL)
{
  os << "Integralgleichung mit folgenden Funktionen:" << endl;
  os << "Kern (Taylorentwicklung):" << endl;
  os << IGL.Kern << endl;
  os << "Restkern:" << endl;
  os << IGL.Restkern << endl;
  os << "Rechte Seite (Taylorentwicklung):" << endl;
  os << IGL.RS << endl;
  return os;
}

