// ***************************************************************************
//
// VFIS Integral Equation Solver (Single Version).
//
// Verified Fredholm Integral Equation Solver
// for linear Fredholm Integral Equations of the second kind
// (Single Version)
//
// Verifizierter Integralgleichungsloeser 
// fuer lineare Fredholmsche Integralgleichungen zweiter Art
// (Einzelversion)
//
// Copyright (c) 2004-2007 Markus Grimmer
//
// markus.grimmer@math.uni-wuppertal.de
//
// VFIS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// VFIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with VFIS; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
// ***************************************************************************


/**
 * @file Einzel-FIGL-Loeser.hpp
 * VFIS Integral Equation Solver central module header with class definitions.
 * Method/Function definitions see implementation file.
 */

#include <iostream>
#include <vector>
#include <string>

using namespace std;
using namespace cxsc;
using namespace taylor;

/**
 * @class IntegralOperatorRG
 * Integral Operator specialized for the Taylor expansion remainder terms 
 * ("Restglied" in German, that's where "RG" comes from...)
 *
 * Application of an integral operator K to a function y: 
 *
 * (Ky)(s) = Int_a^b k(s,t)y(t)dt
 */
class IntegralOperatorRG
{
  private:
    /** 
     * Kernel coefficients from remainder denotation of the kernel function.
     */
    ivector K;
  public:
    /** 
     * Constructor
     * @param iv A C-XSC ivector
     */
    IntegralOperatorRG(ivector& iv) { K=iv; }
    /** 
     * Copy Constructor
     * @param IO An IntegralOperatorRG
     */
    IntegralOperatorRG(const IntegralOperatorRG& IO) { K=IO.K; }
    itaylor operator()(itaylor&, int, interval, ivector&, ivector&);
};

// ---------------------------------------------------------------------------

/**
 * @class IntegralGleichung 
 * The integral equation including all 
 * directly associated data and functions as attributes and methods.
 *
 * Not included: Indirectly associated functions (iteration functions, etc.)
 */
class IntegralGleichung
{
  private:
    /**
     * Typedef for kernel function pointer.
     */
    typedef dim2taylor (*KFkt)(dim2taylor_vector&);
    /**
     * Typedef for right hand side function pointer.
     */
    typedef itaylor (*ReFkt)(itaylor&);
    /**
     * Kernel function pointer.
     */
    KFkt       KernFkt;
    /**
     * Right hand side function pointer.
     */
    ReFkt      RechtsFkt;
    /**
     * @name Kernel Taylor coefficients
     */
    //@{
    dim2taylor Kpoint;
    dim2taylor Kint;
    dim2taylor Kern;
    //@}
    /**
     * @name Right hand side Taylor coefficients
     */
    //@{
    itaylor    RSpoint;
    itaylor    RSint;
    itaylor    RS;  
    //@}
    /**
     * @name Domains and domain midpoints
     */
    //@{
    interval   S;
    interval   T;
    interval   RS_S;
    interval   Smid;
    interval   Tmid;
    interval   RS_Smid;
    //@}
    /**
     * Linear system solution.
     */
    itaylor    ILGS_Lsg;
    /**
     * Integral equation solution.
     */
    itaylor    Lsg;
    /**
     * Taylor order.
     */
    int ord;
    /**
     * Integrals of monomials.
     */
    ivector MonomInt_a;
    /**
     * Integrals of monomials.
     */
    ivector MonomInt_b;
    /**
     * Remainder of kernel Taylor expansion
     */
    ivector Restkern;
  
  public:
  
    IntegralGleichung(int ord=0);
    IntegralGleichung(KFkt,interval, interval,
                      ReFkt, interval, int);
    void Einschliessung_berechnen(ostream& outf=cout);  
    void Loesungsansatz(vector<itaylor>&, itaylor&, int&, ostream& outf);
    void Loesung_bilden(vector<itaylor>&, itaylor&, int&);
    
    /**
     * @name Getters
     * Getters (Read only access functions).
     */
    //@{
    dim2taylor getKern() { return Kern;}
    itaylor getRS() { return RS; }
    itaylor getILGS_Lsg() { return ILGS_Lsg; }
    itaylor getLsg() { return Lsg; }
    interval getS() { return S;} 
    interval getT() { return T;} 
    interval getRS_S() { return RS_S;} 
    //@}
    
    void Export(string dest, string fname, string varname="IGL_Lsg");
    void Export_Plot(string dest, string fname, int steps, 
                     string varname="IGL_Lsg", 
                     string plottitle="SolutionEnclosure");
    
    friend ostream& operator<< (ostream&, IntegralGleichung&);
 
    friend itaylor Iteration (IntegralOperatorRG&, itaylor&, int ord, interval S, 
                              ivector&, ivector&, int& err, ostream& outf);
};
